/*=============================================================================
Vitalnix/ACCDB - Unified Account Database
accdb/rconfig.c - Read simple key-value configuration files
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2004
  -- License restrictions apply (Dual FAL/LGPL2)
  -- For details see doc/FAL.txt and doc/LGPL2.txt.
  $Id: rconfig.c,v 1.7 2004/04/11 09:29:02 hirogen2 Exp $
=============================================================================*/
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#define MAXFNLEN 512  /* max length for filename buffer */
#define MAXLNLEN 1024 /* max length for line buffer */

typedef unsigned char byte;

/*
Multiple possibilities for rconfig_pv()'s flags:

RCONFIG_ONE
  Parsing will stop after one file has been successfully parsed. This allows
  for a "hardcoded+personal" configuration style. If you had

    const char *pv[] = {".", "/usr/local/etc", "/etc", NULL};
    rconfig_pv(pv, "app.conf", table, flags | RCONFIG_ONE);

  only the first (from left to right) successfully parsed will be read.

  To build the current options from default files "hardcoded+default+personal",
  this would be the answer:

    const char *pv[] = {"/etc", "/usr/local/etc", ".", NULL};
    rconfig_pv(pv, "app.conf", table, flags);

  which reads all files, from left to right.
*/

//-----------------------------------------------------------------------------
int accdb_rconfig(const char *file, struct rconfig_opt *tab) {
    char buf[MAXLNLEN];
    FILE *fp = fopen(file, "r");
    if(fp == NULL) { return -errno; }

    while(fgets(buf, MAXLNLEN, fp) != NULL) {
        struct rconfig_opt *travp = tab;
        char *qkey = buf, *qval, *ws;

        HX_chomp(buf);
        // Next entry if comment, empty line or no value
        if(*buf == '#' || *buf == '\0' || (qval = strchr(buf, '=')) == NULL) {
            continue;
        }

        // Skip any whitespace directly before and after '='
        if((ws = strchr(buf, ' ')) != NULL && ws < qval) { *ws = '\0'; }
        else { *qval = '\0'; }
        ++qval;
        while(isspace(*qval)) { ++qval; }

        while(travp->key != NULL) {
            if(strcmp(qkey, travp->key) == 0) {
                switch(travp->type) {
                    case RCONF_BYTE:
                        *(byte *)travp->ptr = *qval;
                        break;
                    case RCONF_UCHAR:
                        *(unsigned char *)travp->ptr = strtol(qval, NULL, 0);
                        break;
                    case RCONF_CHAR:
                        *(char *)travp->ptr = strtol(qval, NULL, 0);
                        break;
                    case RCONF_USHORT:
                        *(unsigned short *)travp->ptr = strtoul(qval, NULL, 0);
                        break;
                    case RCONF_SHORT:
                        *(short *)travp->ptr = strtol(qval, NULL, 0);
                        break;
                    case RCONF_UINT:
                        *(unsigned int *)travp->ptr = strtoul(qval, NULL, 0);
                        break;
                    case RCONF_INT:
                        *(int *)travp->ptr = strtol(qval, NULL, 0);
                        break;
                    case RCONF_ULONG:
                        *(unsigned long *)travp->ptr = strtoul(qval, NULL, 0);
                        break;
                    case RCONF_LONG:
                        *(long *)travp->ptr = strtol(qval, NULL, 0);
                        break;
                    case RCONF_FLOAT:
                        *(float *)travp->ptr = strtod(qval, NULL);
                        break;
                    case RCONF_DOUBLE:
                        *(double *)travp->ptr = strtod(qval, NULL);
                        break;
                    case RCONF_BOOL:
                        if(strcmp(qval, "yes") == 0 ||
                         strcmp(qval, "1") == 0 || strcmp(qval, "on") == 0) {
                            *(byte *)travp->ptr = 1;
                        } else {
                            *(byte *)travp->ptr = 0;
                        }
                        break;
                    case RCONF_IBOOL:
                        if(strcmp(qval, "yes") == 0 ||
                         strcmp(qval, "1") == 0 || strcmp(qval, "on") == 0) {
                            *(int *)travp->ptr = 1;
                        } else {
                            *(int *)travp->ptr = 0;
                        }
                        break;
                    case RCONF_STRING:
                        *(char **)travp->ptr = HX_strdup(qval);
                        break;
                    case RCONF_CB:
                        travp->ptr = qval;
                        break;
                    case RCONF_ULLONG:
                        *(unsigned long *)travp->ptr = strtoull(qval, NULL, 0);
                        break;
                    case RCONF_LLONG:
                        *(long *)travp->ptr = strtoll(qval, NULL, 0);
                        break;
                }

                if(travp->callback != NULL) {
                    travp->callback(travp->key, travp->type, travp->ptr,
                     travp->uptr);
                }
            }
            ++travp;
        }
    }

    fclose(fp);
    return 1;
}

int accdb_rconfig_pv(const char **path, const char *file,
 struct rconfig_opt *tab, unsigned long flags) {
    char buf[MAXFNLEN];
    int rv = 0;

    while(*path != NULL) {
        int v;
        snprintf(buf, MAXLNLEN, "%s/%s", *path, file);
        v = accdb_rconfig(buf, tab);
        if(v > 0) {
            ++rv;
            if(flags & RCONFIG_ONE) { break; }
        }
        ++path;
    }

    return rv;
}

void accdb_rconfig_free(struct rconfig_opt *tab) {
    while(tab->key != NULL) {
        char **ptr = tab->ptr;
        if(tab->type == RCONF_STRING && *ptr != NULL) { free(*ptr); }
        ++tab;
    }
    return;
}

//==[ End of file ]============================================================
