/*=============================================================================
Vitalnix/ACCDB - Unified Account Database
backends/shadow.c - Shadow back-end module
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2002 - 2004
  -- License restrictions apply (Dual FAL/LGPL2)
  -- For details see doc/FAL.txt and doc/LGPL2.txt.
  $Id: shadow.c,v 1.13 2004/04/25 17:20:10 hirogen2 Exp $
=============================================================================*/
#include <sys/file.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libHX.h>
#include <accdb.h>
#include <accdb_int.h>
#define MAXLNLEN 1024 /* max length for line buffer */

#define RETAIN_EMPTY_FIELDS 0

typedef unsigned char byte;

enum {
    ID_USER    = 1 << 0,
    ID_GROUP   = 1 << 1,
    IDF_SYSTEM = 1 << 2,
    CONF_READ  = 0,
    CONF_FREE,
    FORCE_FLUSH = 1 << 0,
};

enum lock_request {
    LOCK_OFF = 0,
    LOCK_ON,
};

struct spdb_file {
    char *path;
    FILE *fp;
    byte is_alloc, is_chg;
};

struct state {
    unsigned long flags;
    long uid_min, uid_max, gid_min, gid_max;
    struct HXlld *user, *group;
    struct HXlld_node *utrav, *gtrav;
    struct spdb_file fpasswd, fshadow, fgroup, fgshadow;
};

static void db_flush__unlock(struct state *);
static int db_flush(struct state *, int);
static long find_next_id(struct state *, long, unsigned long);
inline static void free_single_group(struct accdb_group *);
static void free_groups(struct state *);
inline static void free_single_user(struct accdb_user *);
static void free_users(struct state *);
static struct accdb_group *get_group(struct state *, char *);
static struct accdb_user *get_user(struct state *, char *);
static int lock_file(FILE *, enum lock_request);
static void read_config(struct state *, int);
static int pconfig__ack(const char *, byte, void *, void *);
static int read_groups(struct state *);
static int read_users(struct state *);

MODULE_NAME("Shadow back-end module");
MODULE_DESC("for /etc/passwd and shadow");
MODULE_INFO("Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2002 - 2004");

static char *def_fpasswd_path = "/etc/passwd",
 *def_fshadow_path = "/etc/shadow", *def_fgroup_path = "/etc/group",
 *def_fgshadow_path = "/etc/gshadow";

//-----------------------------------------------------------------------------
int q_open(struct accdb_module *mp) {
    struct state *sp;
    int eax;

    if((sp = mp->state = calloc(1, sizeof(struct state))) == NULL) {
        return -errno;
    }

    sp->fpasswd.path = def_fpasswd_path;
    sp->fshadow.path = def_fshadow_path;
    sp->fgroup.path = def_fgroup_path;
    sp->fgshadow.path = def_fgshadow_path;
    read_config(sp, CONF_READ);

    if((sp->fpasswd.fp = fopen(sp->fpasswd.path, "r+")) == NULL ||
     (sp->fshadow.fp = fopen(sp->fshadow.path, "r+")) == NULL ||
     (sp->fgroup.fp = fopen(sp->fgroup.path, "r+")) == NULL ||
     (sp->fgshadow.fp = fopen(sp->fgshadow.path, "r+")) == NULL) {
        // Could not open at least one of the DB core files
        eax = errno;
        q_close(mp);
        return -(errno = eax);
    }

    if((eax = read_users(sp)) < 0 || (eax = read_groups(sp)) < 0) {
        // Error while reading users/groups
        eax = errno;
        q_close(mp);
        return -(errno = eax);
    }

    sp->utrav = sp->user->first;
    sp->gtrav = sp->group->first;
    return 1;
}

//-----------------------------------------------------------------------------
int q_usertrav(struct accdb_module *mp, struct accdb_user *result) {
    struct state *sp = mp->state;
    struct accdb_user *src;
    struct accdb_group qgrp = {gname: NULL};

    if(result == NULL) {
        sp->utrav = sp->user->first;
        return 0;
    }

    if(sp->utrav == NULL) { return 0; }
    src = sp->utrav->ptr;

    while(*src->lname == '+' || *src->lname == '-') {
        // skip NIS users
        if((sp->utrav = sp->utrav->Next) == NULL) { return 0; }
        src = sp->utrav->ptr;
    }

    memcpy(result, src, sizeof(struct accdb_user));
    qgrp.gid = src->gid;
    q_groupinfo(mp, &qgrp, &qgrp, 1);
    result->igrp = qgrp.gname;

    sp->utrav = sp->utrav->Next;
    return 1;
}

int q_userinfo(struct accdb_module *mp, struct accdb_user *rq,
 struct accdb_user *dest, size_t s) {
    struct state *sp = mp->state;
    struct HXlld_node *travp = sp->user->first;
    struct accdb_user xrq;
    int found = 0;

    if(rq == dest) {
        memcpy(&xrq, rq, sizeof(struct accdb_user));
        rq = &xrq;
    }

    while(travp != NULL && (dest == NULL || (dest != NULL && s > 0))) {
        struct accdb_user *src = travp->ptr;

        // happily skip NIS users
        if(*src->lname == '+' || *src->lname == '-' ||
         (rq->lname != NULL && strcmp(rq->lname, src->lname) != 0) ||
         (rq->uid != -1 && rq->uid != src->uid) ||
         (rq->gid != -1 && rq->gid != src->gid)) {
            travp = travp->Next;
            continue;
        }

      if(dest != NULL) {
          struct accdb_group qgrp = {gid: src->gid, gname: NULL};
          memcpy(dest, src, sizeof(struct accdb_user));
          q_groupinfo(mp, &qgrp, &qgrp, 1);
          dest->igrp = qgrp.gname;
          ++dest;
          ++found;
          --s;
      } else {
          if(s == 0) { return 1; }
          ++found;
      }
      travp = travp->Next;
    }

    return found;
}

int q_useradd(struct accdb_module *mp, struct accdb_user *u) {
    struct state *sp = mp->state;
    struct accdb_group qgrp;
    struct accdb_user *nu;
    int eax;

    if(u->lname == NULL || (u->gid == -1 && u->igrp == NULL)) {
        errno = EINVAL;
        return 0;
    }

    qgrp.gid   = u->gid;
    qgrp.gname = u->igrp;
    eax = q_groupinfo(mp, &qgrp, &qgrp, 1);
    if(eax < 0) { return -errno; }
    if(eax == 0) {
        errno = ENOENT;
        return 0;
    }

    /* We have to be careful with strings here, as we need to duplicate them
    for us, leave the original ones intact and not have a single foreign
    pointer in our structs. */
    if((nu = calloc(1, sizeof(struct accdb_user))) == NULL) { return -errno; }

    nu->lname = HX_strdup(u->lname);
    nu->uid   = u->uid;
    if(nu->uid == -1) { nu->uid = find_next_id(sp, -1, ID_USER); }
    nu->gid   = qgrp.gid;
    nu->gecos = HX_strdup(u->gecos);
    nu->home  = HX_strdup(u->home);
    nu->shell = HX_strdup(u->shell);
    nu->igrp  = qgrp.gname;
    nu->sgrp  = HX_strdup(u->sgrp);

    nu->passwd      = HX_strdup(u->passwd);
    nu->last_change = u->last_change;
    nu->keep_min    = u->keep_min;
    nu->keep_max    = u->keep_max;
    nu->warn_age    = u->warn_age;
    nu->expire      = u->expire;
    nu->inactive    = u->inactive;

    // We leave it up to the application to check for duplicates
    HXlld_push(sp->user, nu);
    sp->fpasswd.is_chg = sp->fshadow.is_chg = 1;
    if(db_flush(sp, 0) <= 0) { return -errno; }

    // Update info struct
    u->uid = nu->uid;
    u->gid = nu->gid;
    return 1;
}

int q_usermod(struct accdb_module *mp, struct accdb_user *sm,
 struct accdb_user *mm) {
#define UP_INT(__field) \
    if(mm->__field != -1) { sm->__field = mm->__field; }
#define UP_STR(__field) \
    if(mm->__field != NULL) { HX_strclone(&sm->__field, mm->__field); }
    struct state *sp = mp->state;

    // XXX: Reduce this overhead (searching user list twice)
    if(q_userinfo(mp, sm, NULL, 0) <= 0 ||
     (sm = get_user(sp, sm->lname)) == NULL) {
	errno = ENOENT;
	return 0;
    }

    struct accdb_group qgrp = {gname: sm->igrp, gid: sm->gid};
    q_groupinfo(mp, &qgrp, &qgrp, 1);

    UP_STR(lname);
    UP_INT(uid);
    UP_INT(gid);
    UP_STR(gecos);
    UP_STR(home);
    UP_STR(shell);
    UP_STR(igrp);
    UP_STR(sgrp);

    UP_STR(passwd);
    UP_INT(last_change);
    UP_INT(keep_min);
    UP_INT(keep_max);
    UP_INT(warn_age);

    // XXX: cannot be updated to '-1'
    UP_INT(expire);
    UP_INT(inactive);

    sp->fpasswd.is_chg = sp->fshadow.is_chg = 1;
    if(db_flush(sp, 0) <= 0) { return -errno; }
    return 1;
#undef UP_INT
#undef UP_STR
}

int q_userdel(struct accdb_module *mp, struct accdb_user *u) {
    struct state *sp = mp->state;
    if((u = get_user(sp, u->lname)) == NULL) {
	errno = ENOENT;
	return 0;
    }

    HXlld_del(HXlld_find(sp->user, u));
    free_single_user(u);
    sp->fpasswd.is_chg = sp->fshadow.is_chg = 1;
    if(db_flush(sp, 0) <= 0) { return -errno; }
    return 1;
}

//-----------------------------------------------------------------------------
int q_grouptrav(struct accdb_module *mp, struct accdb_group *result) {
    struct state *sp = mp->state;
    struct accdb_group *x;

    if(result == NULL) {
        sp->gtrav = sp->group->first;
        return 1;
    }

    if(sp->gtrav == NULL) { return 0; }
    x = sp->gtrav->ptr;

    while(*x->gname == '+' || *x->gname == '-') {
        // skip NIS groups
        if((sp->gtrav = sp->gtrav->Next) == NULL) { return 0; }
        x = sp->gtrav->ptr;
    }

    memcpy(result, x, sizeof(struct accdb_group));
    sp->gtrav = sp->gtrav->Next;
    return 1;
}

int q_groupinfo(struct accdb_module *mp, struct accdb_group *rq,
 struct accdb_group *dest, size_t s) {
    struct state *sp = mp->state;
    struct HXlld_node *travp = sp->group->first;
    struct accdb_group xrq;
    int found = 0;

    if(rq == dest) {
        memcpy(&xrq, rq, sizeof(struct accdb_group));
        rq = &xrq;
    }

    while(travp != NULL && (dest == NULL || (dest != NULL && s > 0))) {
        struct accdb_group *i = travp->ptr;

        // happily skip NIS groups
        if(*i->gname == '+' || *i->gname == '-' ||
         (rq->gid != -1 && rq->gid != i->gid) ||
         (rq->gname != NULL && strcmp(rq->gname, i->gname) != 0)) {
            travp = travp->Next;
            continue;
        }

        if(dest != NULL) {
            memcpy(dest, i, sizeof(struct accdb_group));
            ++dest;
            ++found;
            --s;
        } else {
            return 1;
        }
        travp = travp->Next;
    }

    return found;
}

int q_groupadd(struct accdb_module *mp, struct accdb_group *g) {
    struct state *sp = mp->state;
    struct accdb_group *ng;

    if(g->gname == NULL) {
        errno = EINVAL;
        return 0;
    }

    if((ng = calloc(1, sizeof(struct accdb_group))) == NULL) { return -errno; }
    ng->gname = HX_strdup(g->gname);
    ng->gid = g->gid;
    if(ng->gid == -1) { ng->gid = find_next_id(sp, -1, ID_GROUP); }

    HXlld_push(sp->group, ng);
    sp->fgroup.is_chg = sp->fgshadow.is_chg = 1;
    if(db_flush(sp, 0) <= 0) { return -errno; }

    // Update info struct
    g->gid = ng->gid;
    return 1;
}

int q_groupmod(struct accdb_module *mp, struct accdb_group *sm,
 struct accdb_group *mm) {
    struct state *sp = mp->state;

    /* We first check if the group in SM (search mask) is available, then
    replace SM to point to our internal data structures. */
    if(q_groupinfo(mp, sm, NULL, 0) <= 0 ||
     (sm = get_group(sp, sm->gname)) == NULL) {
        errno = ENOENT;
        return 0;
    }

    if(mm->gname != NULL) { HX_strclone(&sm->gname, mm->gname); }
    if(mm->gid != -1) { sm->gid = mm->gid; }

    sp->fgroup.is_chg = 1;
    if(db_flush(sp, 0) <= 0) { return -errno; }
    return 1;
}

int q_groupdel(struct accdb_module *mp, struct accdb_group *g) {
    struct state *sp = mp->state;
    if((g = get_group(sp, g->gname)) == NULL) {
        errno = ENOENT;
        return 0;
    }

    HXlld_del(HXlld_find(sp->group, g));
    free_single_group(g);
    sp->fgroup.is_chg = sp->fgshadow.is_chg = 1;
    if(db_flush(sp, 0) <= 0) { return -errno; }
    return 1;
}

//-----------------------------------------------------------------------------
int q_close(struct accdb_module *mp) {
#define fclose_null(fp) /* guard against NULL */ \
  if((fp) != NULL) { fclose(fp); }
    struct state *sp = mp->state;
    db_flush(sp, FORCE_FLUSH);
    fclose_null(sp->fpasswd.fp);
    fclose_null(sp->fshadow.fp);
    fclose_null(sp->fgroup.fp);
    fclose_null(sp->fgshadow.fp);
    free_users(sp);
    free_groups(sp);
    read_config(sp, CONF_FREE);
    free(sp);
    return 1;
#undef fclose_null
}

long q_sysctl(unsigned int req, ...) {
    enum { // accdb_shadow specific requests
        ACCDB_CHDB_PASSWD = 0xC000,
        ACCDB_CHDB_SHADOW,
        ACCDB_CHDB_GROUP,
        ACCDB_CHDB_GSHADOW,
    };

    struct state *sp = NULL;
    int rv = 1;
    va_list argp;
    va_start(argp, req);

    switch(req) {
        case ACCDB_ADDFLAGS:
            /* q_sysctl(unsigned int req, struct accdb_module *mp,
                 unsigned long fl); */
            sp = va_arg(argp, struct accdb_module *)->state;
            rv = sp->flags |= va_arg(argp, unsigned long);
            break;
        case ACCDB_DELFLAGS:
            /* q_sysctl(unsigned int req, struct accdb_module *mp,
                 unsigned long fl); */
            sp = va_arg(argp, struct accdb_module *)->state;
            rv = sp->flags &= ~va_arg(argp, unsigned long);
            break;
        case ACCDB_FLUSHDB:
            // q_sysctl(unsigned int req, struct accdb_module *mp);
            db_flush(va_arg(argp, struct accdb_module *)->state, FORCE_FLUSH);
            break;
        case ACCDB_COUNT_USERS:
            // q_sysctl(unsigned int req, struct accdb_module *mp);
            sp = va_arg(argp, struct accdb_module *)->state;
            rv = sp->user->itemcount;
            break;
        case ACCDB_COUNT_GROUPS:
            // q_sysctl(unsigned int req, struct accdb_module *mp);
            sp = va_arg(argp, struct accdb_module *)->state;
            rv = sp->group->itemcount;
            break;

        case ACCDB_NEXTUID_SYS:
            // q_sysctl(unsigned int req, struct accdb_module *mp);
            rv = find_next_id(va_arg(argp, struct accdb_module *)->state,
             -1, ID_USER | IDF_SYSTEM);
            break;
        case ACCDB_NEXTUID:
            // q_sysctl(unsigned int req, struct accdb_module *sp);
            rv = find_next_id(va_arg(argp, struct accdb_module *)->state,
             -1, ID_USER);
            break;
        case ACCDB_NEXTGID_SYS:
            // q_sysctl(unsigned int req, struct accdb_module *sp);
            rv = find_next_id(va_arg(argp, struct accdb_module *)->state,
             -1, ID_GROUP | IDF_SYSTEM);
            break;
        case ACCDB_NEXTGID:
            // q_sysctl(unsigned int req, struct accdb_module *sp);
            rv = find_next_id(va_arg(argp, struct accdb_module *)->state,
             -1, ID_GROUP);
            break;

        // Special handling for SLANED
        case ACCDB_LOCK_EXCL:
            break;
        case ACCDB_UNLOCK_EXCL:
            break;

        // These internal sysctl requests are only used for debugging
        case ACCDB_CHDB_PASSWD:
            def_fpasswd_path = HX_strdup(va_arg(argp, const char *));
            break;
        case ACCDB_CHDB_SHADOW:
            def_fshadow_path = HX_strdup(va_arg(argp, const char *));
            break;
        case ACCDB_CHDB_GROUP:
            def_fgroup_path = HX_strdup(va_arg(argp, const char *));
            break;
        case ACCDB_CHDB_GSHADOW:
            def_fgshadow_path = HX_strdup(va_arg(argp, const char *));
            break;

        default:
            rv = -ENOSYS;
            break;
    }

    va_end(argp);
    return rv;
}

//-----------------------------------------------------------------------------
static void db_flush__unlock(struct state *sp) {
    lock_file(sp->fpasswd.fp,  LOCK_OFF);
    lock_file(sp->fshadow.fp,  LOCK_OFF);
    lock_file(sp->fgroup.fp,   LOCK_OFF);
    lock_file(sp->fgshadow.fp, LOCK_OFF);
    return;
}

static int db_flush(struct state *sp, int force) {
    int eax;
    if(!force && !(sp->flags & ACCDB_AFLUSH)) { return 1; }

    /* Locks the corresponding database files and writes all users stored in
    the state back to disk. */
    if((eax = lock_file(sp->fpasswd.fp, LOCK_ON)) < 0 ||
     (eax = lock_file(sp->fshadow.fp,   LOCK_ON)) < 0 ||
     (eax = lock_file(sp->fgroup.fp,    LOCK_ON)) < 0 ||
     (eax = lock_file(sp->fgshadow.fp,  LOCK_ON)) < 0) {
        db_flush__unlock(sp);
        return eax;
    }

    if(sp->fpasswd.is_chg || sp->fshadow.is_chg) {
        // /etc/passwd and /etc/shadow
        struct HXlld_node *travp = sp->user->first;

#ifndef _WIN32
        // Some weird operations to get the desired effect
        fseek(sp->fpasswd.fp, 0, SEEK_SET);
        fseek(sp->fshadow.fp, 0, SEEK_SET);
        fflush(sp->fpasswd.fp);
        fflush(sp->fshadow.fp);
        ftruncate(fileno(sp->fpasswd.fp), 0);
        ftruncate(fileno(sp->fshadow.fp), 0);
        fflush(sp->fpasswd.fp);
        fflush(sp->fshadow.fp);
#else
        /* No simple locking on Win32 without modifying more code, so sticking
        with something very plain. */
        freopen(sp->fpasswd.path, "w", sp->fpasswd.fp);
        freopen(sp->fshadow.path, "w", sp->fshadow.fp);
#endif

        while(travp != NULL) {
            struct accdb_user *u = travp->ptr;
            char **pr = u->_private;
            char *password = (u->passwd != NULL) ? u->passwd : "!";

            fprintf(sp->fpasswd.fp, "%s:x:%ld:%ld:%s:%s:%s",
             u->lname, u->uid, u->gid, u->gecos, u->home, u->shell);
            if(pr != NULL && pr[0] != NULL) {
                fprintf(sp->fpasswd.fp, "%s", pr[0]);
            }
            fprintf(sp->fpasswd.fp, "\n");

#if RETAIN_EMPTY_FIELDS
            fprintf(sp->fshadow.fp, "%s:%s:%ld:%ld:%ld:%ld:",
             u->lname, password, u->last_change, u->keep_min, u->keep_max,
             u->warn_age);
            if(u->expire >= 0) {
                fprintf(sp->fshadow.fp, "%ld", u->expire);
            }
            fprintf(sp->fshadow.fp, ":");
            if(u->inactive >= 0) {
                fprintf(sp->fshadow.fp, "%ld", u->inactive);
            }
            fprintf(sp->fshadow.fp, ":");
            if(pr != NULL && pr[1] != NULL) {
                fprintf(sp->fshadow.fp, "%s", pr[1]);
            }
#else
            /* Shadow is compatibile with -1 in .expire and .inactive, so use
            this slightly faster approach. */
            fprintf(sp->fshadow.fp, "%s:%s:%ld:%ld:%ld:%ld:%ld:%ld:",
             u->lname, password, u->last_change, u->keep_min, u->keep_max,
             u->warn_age, u->expire, u->inactive);
            if(pr != NULL && pr[1] != NULL) {
                fprintf(sp->fshadow.fp, "%s", pr[1]);
            }
#endif
            fprintf(sp->fshadow.fp, "\n");

            travp = travp->Next;
        }

        sp->fpasswd.is_chg = sp->fshadow.is_chg = 0;
    }

    if(sp->fgroup.is_chg || sp->fgshadow.is_chg) {
        // /etc/group and /etc/gshadow
        struct HXlld_node *travp = sp->group->first;

#ifndef _WIN32
        fseek(sp->fgroup.fp,   0, SEEK_SET);
        fseek(sp->fgshadow.fp, 0, SEEK_SET);
        fflush(sp->fgroup.fp);
        fflush(sp->fgshadow.fp);
        ftruncate(fileno(sp->fgroup.fp),   0);
        ftruncate(fileno(sp->fgshadow.fp), 0);
        fflush(sp->fgroup.fp);
        fflush(sp->fgshadow.fp);
#else
        freopen(sp->fgroup.path,   "w", sp->fgroup.fp);
        freopen(sp->fgshadow.path, "w", sp->fgshadow.fp);
#endif

        while(travp != NULL) {
            struct accdb_group *g = travp->ptr;
            char **pr = g->_private;
            fprintf(sp->fgroup.fp, "%s:x:%ld:", g->gname, g->gid);
            if(pr != NULL && pr[0] != NULL) {
                fprintf(sp->fgroup.fp, "%s", pr[0]);
            }
            fprintf(sp->fgroup.fp, "\n");

            fprintf(sp->fgshadow.fp, "%s:", g->gname);
            if(pr != NULL && pr[1] != NULL) {
                fprintf(sp->fgshadow.fp, "%s", pr[1]);
            } else {
                fprintf(sp->fgshadow.fp, "*:root:");
            }
            fprintf(sp->fgshadow.fp, "\n");
            travp = travp->Next;
        }

        sp->fgroup.is_chg = sp->fgshadow.is_chg = 0;
    }

    db_flush__unlock(sp);
    return 1;
}

static long find_next_id(struct state *sp, long want, unsigned long flags) {
    long min, max;

    if(flags & ID_USER) {
        /* Tries to find the next free UID. If u is >= 0, that specific UID is
        wanted. In that case, u is returned if it is free, or -EEXIST
        otherwise. For u < 0 (just get the next free UID), the next free one is
        returned or -ENOENT if none could be found between sp->uid_min and
        sp->uid_max. */
        struct HXlld_node *travp = sp->user->first;
        byte used = 0;

        if(flags & IDF_SYSTEM) {
            min = 1;
            max = sp->uid_min - 1;
        } else {
            min = sp->uid_min;
            max = sp->uid_max;
        }

        if(want >= 0) {
            while(travp != NULL) {
                if(((struct accdb_user *)travp->ptr)->uid == want) {
                    ++used;
                    break;
                }
                travp = travp->Next;
            }
            if(used) { return -(errno = EEXIST); }
            return want;
        }

        /* No specific UID: As the original "useradd" from the Shadow package
        did, pick the UID after the newest user below or equal to UID_MAX. */
        want = min;
        while(travp != NULL) {
            long this = ((struct accdb_user *)travp->ptr)->uid;
            if(this >= want && this <= max) {
              if(want == max) { want = -1; break; }
              want = this + 1;
            }
            travp = travp->Next;
        }

        if(want >= 0) { return want; }

        /* There is no more free UID after the newest user <= UID_MAX, fall
        back to method #2: get the next free UID, starting with UID_MIN. */
        want = min;
        while(want <= max) {
            travp = sp->user->first;
            while(travp != NULL) {
                if(((struct accdb_user *)travp->ptr)->uid == want) {
                    ++used;
                    break;
                }
                travp = travp->Next;
            }
            if(!used) { return want; }
            --used;
            ++want;
        }

        return -(errno = ENOENT);
    } else if(flags & ID_GROUP) {
        // Same for groups
        struct HXlld_node *travp = sp->group->first;
        byte used = 0;

        if(flags & IDF_SYSTEM) {
            min = 1;
            max = sp->gid_min - 1;
        } else {
            min = sp->gid_min;
            max = sp->gid_max;
        }

        if(want >= 0) {
            while(travp != NULL) {
                if(((struct accdb_user *)travp->ptr)->gid == want) {
                    ++used;
                    break;
                }
                travp = travp->Next;
            }
            if(used) { return -(errno = EEXIST); }
            return want;
        }

        want = min;
        while(travp != NULL) {
            long this = ((struct accdb_group *)travp->ptr)->gid;
            if(this >= want && this <= max) {
                if(want == max) { want = -1; break; }
                want = this + 1;
            }
            travp = travp->Next;
        }

        if(want >= 0) { return want; }

        want = min;
        while(want <= max) {
            travp = sp->group->first;
            while(travp != NULL) {
                if(((struct accdb_user *)travp->ptr)->uid == want) {
                    ++used;
                    break;
                }
                travp = travp->Next;
            }
            if(!used) { return want; }
            --used;
            ++want;
        }
        return -(errno = ENOENT);
    }

    return -1;
}

inline static void free_single_user(struct accdb_user *u) {
    char **p = u->_private;
    free(u->lname);
    free(u->gecos);
    free(u->home);
    free(u->shell);
    free(u->sgrp);
    free(u->passwd);

    if(p != NULL) {
        free(p[0]);
        free(p[1]);
        free(p);
    }

    free(u);
    return;
}

static void free_users(struct state *sp) {
    struct HXlld_node *travp;
    if(sp->user == NULL) { return; }
    travp = sp->user->first;

    while(travp != NULL) {
        free_single_user(travp->ptr);
        travp = travp->Next;
    }

    HXlld_free(sp->user);
    return;
}

inline static void free_single_group(struct accdb_group *g) {
    char **p = g->_private;
    free(g->gname);
    if(p != NULL) {
        free(p[0]);
        free(p[1]);
        free(p);
    }
    free(g);
    return;
}

static void free_groups(struct state *sp) {
    struct HXlld_node *travp;
    if(sp->group == NULL) { return; }
    travp = sp->group->first;

    while(travp != NULL) {
        free_single_group(travp->ptr);
        travp = travp->Next;
    }

    HXlld_free(sp->group);
    return;
}

static struct accdb_user *get_user(struct state *sp, char *lname) {
    struct HXlld_node *travp = sp->user->first;

    while(travp != NULL) {
        struct accdb_user *u = travp->ptr;
        if(strcmp(u->lname, lname) == 0) { return u; }
        travp = travp->Next;
    }

    return NULL;
}

static struct accdb_group *get_group(struct state *sp, char *name) {
    // XXX....this looks like a faster function as q_userinfo()...
    // but then again, only compares against names. Keep it?
    struct HXlld_node *travp = sp->group->first;

    while(travp != NULL) {
        struct accdb_group *g = travp->ptr;
        if(strcmp(g->gname, name) == 0) { return g; }
        travp = travp->Next;
    }

    return NULL;
}

static int lock_file(FILE *fp, enum lock_request req) {
    int rv = -1, eax = 0;
    long pos;

    if(fp == NULL) { return -EINVAL; }
    pos = ftell(fp);
    fseek(fp, 0, SEEK_SET);
    fflush(fp);

    switch(req) {
#ifndef _WIN32
        case LOCK_OFF:
            rv = lockf(fileno(fp), F_ULOCK, 0);
            eax = errno;
            break;
        case LOCK_ON:
            rv = lockf(fileno(fp), F_LOCK, 0);
            eax = errno;
            break;
#else
        case LOCK_OFF:
        case LOCK_ON:
            rv = 0;
            break;
#endif
    }

    fseek(fp, pos, SEEK_SET);
    fflush(fp);
    errno = eax;
    if(rv == -1) { return -errno; }
    return 1;
}

static void read_config(struct state *sp, int f) {
    const char *paths[] = {"backends/../vetc", "/etc", NULL};
    struct rconfig_opt autouid_table[] = {
        {"UID_MIN", RCONF_LONG, &sp->uid_min, NULL, NULL},
        {"UID_MAX", RCONF_LONG, &sp->uid_max, NULL, NULL},
        {"GID_MIN", RCONF_LONG, &sp->gid_min, NULL, NULL},
        {"GID_MAX", RCONF_LONG, &sp->gid_max, NULL, NULL},
        {NULL},
    };
    struct rconfig_opt options_table[] = {
        {"PASSWD_DB",  RCONF_STRING, &sp->fpasswd.path,
         pconfig__ack, &sp->fpasswd.is_alloc},
        {"SHADOW_DB",  RCONF_STRING, &sp->fshadow.path,
         pconfig__ack, &sp->fshadow.is_alloc},
        {"GROUP_DB",   RCONF_STRING, &sp->fgroup.path,
         pconfig__ack, &sp->fgroup.is_alloc},
        {"GSHADOW_DB", RCONF_STRING, &sp->fgshadow.path,
         pconfig__ack, &sp->fgshadow.is_alloc},
        {NULL},
    };

    if(f == CONF_FREE) {
        accdb_rconfig_free(options_table);
    } else {
        sp->uid_min = sp->gid_min = 100;
        sp->uid_max = sp->gid_max = 60000;

        accdb_rconfig_pv(paths, "autouid", autouid_table, RCONFIG_ONE);
        accdb_rconfig_pv(paths, "libaccdb/shadow", options_table, RCONFIG_ONE);
    }

    return;
}

static int pconfig__ack(const char *key, byte fmt, void *ptr, void *uptr) {
    *(int *)uptr = 1;
    return 1;
}

static int read_users(struct state *sp) {
    /* In this function, we read all users into memory. This speeds processing
    up, as seeking and scanning in userinfo() / groupinfo() would take long
    otherwise. We also have it handy in a struct. Doing so also allows us to
    commit all changes to the user database at once (actually only the case
    with the ACCDB_NOAUTOFLUSH flag).

    Strings are allocated by us -- and freed by us. The char * pointers the
    application gets are only for reading (and dup'ing). */
    struct accdb_user *u;
    byte buf[MAXLNLEN];
    if((sp->user = HXlld_init(0)) == NULL) { return -errno; }

    while(fgets(buf, MAXLNLEN, sp->fpasswd.fp) != NULL) {
        char *bufp = buf, *pa;
        if(*bufp == '#') { continue; }
        if((u = calloc(1, sizeof(struct accdb_user))) == NULL) { break; }
        HX_chomp(bufp);

        u->lname  = HX_strdup(HX_strsep(&bufp, ":"));
        u->passwd = HX_strsep(&bufp, ":");
        u->uid    = strtol(HX_strsep(&bufp, ":"), NULL, 0);
        u->gid    = strtol(HX_strsep(&bufp, ":"), NULL, 0);
        u->gecos  = HX_strdup(HX_strsep(&bufp, ":"));
        u->home   = HX_strdup(HX_strsep(&bufp, ":"));
        u->shell  = HX_strdup(HX_strsep(&bufp, ":"));

        /* Anything after the seven main fields is "reserved for future use",
        however, we need to keep that intact. */
        u->_private = calloc(1, sizeof(void *) * 2);
        if((pa = HX_strsep(&bufp, ":")) != NULL && *pa != '\0') {
            ((char **)u->_private)[0] = HX_strdup(pa);
        }

        // In case there is no shadow entry, we need some recovery values
        u->keep_min    = 0;
        u->keep_max    = 366;
        u->warn_age    = 21;
        u->last_change = \
        u->expire      = \
        u->inactive    = -1;

        /* Stuff is pushed in a linked list. The list is unordered, so it has
        the same order as /etc/passwd. This is intentional (and suffices). */
        HXlld_push(sp->user, u);
    }

    // Update the sp->user table with the data from /etc/shadow
    while(fgets(buf, MAXLNLEN, sp->fshadow.fp) != NULL) {
        char *bufp = buf, *lname, *pa;
        if(*buf == '#') { continue; }
        HX_chomp(buf);

        lname = HX_strsep(&bufp, ":");
        if((u = get_user(sp, lname)) == NULL) { continue; } // orphaned entry
        u->passwd      = HX_strdup(HX_strsep(&bufp, ":"));
        u->last_change = strtol(HX_strsep(&bufp, ":"), NULL, 0);
        u->keep_min    = strtol(HX_strsep(&bufp, ":"), NULL, 0);
        u->keep_max    = strtol(HX_strsep(&bufp, ":"), NULL, 0);
        u->warn_age    = strtol(HX_strsep(&bufp, ":"), NULL, 0);

        pa = HX_strsep(&bufp, ":");
        if(*pa == '\0') { u->expire = -1; }
        else { u->expire = strtol(pa, NULL, 0); }

        pa = HX_strsep(&bufp, ":");
        if(*pa == '\0') { u->inactive = -1; }
        else { u->inactive = strtol(pa, NULL, 0); }

        // Retain reserved fields
        if((pa = HX_strsep(&bufp, ":")) != NULL && *pa != '\0') {
            ((char **)u->_private)[1] = HX_strdup(pa);
        }
    }

    return 1;
}

static int read_groups(struct state *sp) {
    // Comments to be found in read_users()
    struct accdb_group *g;
    byte buf[MAXLNLEN];
    if((sp->group = HXlld_init(0)) == NULL) { return -errno; }

    while(fgets(buf, MAXLNLEN, sp->fgroup.fp) != NULL) {
        char *bufp = buf, *pa;
        if(*bufp == '#') { continue; }
        if((g = malloc(sizeof(struct accdb_group))) == NULL) { return -errno; }
        HX_chomp(bufp);

        g->gname = HX_strdup(HX_strsep(&bufp, ":"));
        HX_strsep(&bufp, ":");
        g->gid      = strtol(HX_strsep(&bufp, ":"), NULL, 0);
        g->_private = calloc(1, sizeof(void *) * 2);

        if((pa = HX_strsep(&bufp, ":")) != NULL && *pa != '\0') {
            ((char **)g->_private)[0] = HX_strdup(pa);
        }

        HXlld_push(sp->group, g);
    }

    // Update the sp->group table with the data from /etc/gshadow
    while(fgets(buf, MAXLNLEN, sp->fgshadow.fp) != NULL) {
        char *bufp = buf, *name;
        if(*bufp == '#') { continue; }
        HX_chomp(bufp);

        name = HX_strsep(&bufp, ":");
        if((g = get_group(sp, name)) == NULL) { continue; } // orphaned entry
        if(*bufp != '\0') { ((char **)g->_private)[1] = HX_strdup(bufp); }
    }

    return 1;
}

//==[ End of file ]============================================================
