/*=============================================================================
Vitalnix/C-spark - User Management CLI/CUI
data.c - Data processing
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 1999 - 2004
  -- License restrictions apply (Dual FAL/GPL2)
  -- For details see doc/FAL.txt and doc/GPL2.txt.
  $Id: data.c,v 1.7 2004/05/11 13:47:36 hirogen2 Exp $
=============================================================================*/
#include <sys/stat.h>
#include <sys/types.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libHX.h>
#include "cui/global.h"

enum {
    CLASS_NONE = 0,
    CLASS_DIGIT,
    CLASS_ALPHA,
    CLASS_OTHER,
    /* compare_wbc() compares by class, then by context (i.e. numerically for
    CLASS_DIGIT and lexically for CLASS_ALPHA and CLASS_OTHER). The context
    with the smallest enumeration (see above) is considered less - in all
    cases (when two segments are of different class). */
};

static char *transform7(const byte *, char *, size_t);
static int wb_classof(char);

static const struct stab {
    char *set, *out;
} subst_tab[] = {
    {"", "ae"}, {"", "oe"}, {"", "ue"}, {"", "ss"},
    {"", "a"}, {"", "c"}, {"", "e"}, {"", "i"},
    {"", "o"}, {"", "u"}, {"", "l"}, {"", "m"},
    {"", "n"}, {"", "r"}, {"", "y"}, {NULL, NULL},
};

//-----------------------------------------------------------------------------
int compare_wbc(const char *str_a, const char *str_b) {
    /* Split both strings at word boundaries and compare each fragment pair
    according to context (numerical or lexically). (They are not really split,
    but it mimics Perl's regular \b expression.)

    A boundary in compare_wbc() is between two characters which are of
    different class. Classes are: '\0', [0-9], [a-z], [^0-9a-z]. (Thus, the
    following would hold true:
    11C < 12C, 8EK < 8F2, 10A > 9B, 9 < A2, B > A2, 0 < '-') */

    const char *walk_a = str_a, *walk_b = str_b;
    int cl_a, cl_b;

    cl_a = wb_classof(*walk_a++);
    cl_b = wb_classof(*walk_b++);
    if(cl_a != cl_b) { return cl_a - cl_b; }

    /* The clever analyst saw that cl_a == cl_b, thus being able to reduce
    cl_a == CLASS_NONE && cl_b == CLASS_NONE into...: */
    if(cl_a == CLASS_NONE) { return 0; }

    while(1) {
        char buf_a[MAXSNLEN], buf_b[MAXSNLEN];
        int d;

        // Find the next segment start
        while(wb_classof(*walk_a) == cl_a) { ++walk_a; }
        while(wb_classof(*walk_b) == cl_b) { ++walk_b; }

        // Compare current segments' content
        d = walk_a - str_a;
        buf_a[d] = '\0';
        strncpy(buf_a, str_a, d);

        d = walk_b - str_b;
        buf_b[d] = '\0';
        strncpy(buf_b, str_b, d);

        if(cl_a == CLASS_DIGIT) {
            int va = strtoul(buf_a, NULL, 10), vb = strtoul(buf_b, NULL, 10);
            if(va != vb) { return (va > vb) - (va < vb); }
        } else if(cl_a == CLASS_ALPHA || cl_a == CLASS_OTHER) {
            if((d = strcmp(buf_a, buf_b)) != 0) { return d; }
        } else if(cl_a == CLASS_NONE) {
            return 0;
        }

        // Current segment was equal [again], so try next segment.
        cl_a = wb_classof(*walk_a);
        cl_b = wb_classof(*walk_b);
        if(cl_a != cl_b) { return cl_a - cl_b; }
        if(cl_a == CLASS_NONE) { return 0; }

        str_a = walk_a;
        str_b = walk_b;
    }

    return 0;
}

int create_home(const char *d, long uid, long gid) {
    if(HX_mkdir(d) <= 0) { return 0; }
    lchown(d, uid, gid);
    chmod(d, 0711);
    if(Optmap.f_skeld != NULL) {
        HX_copy_dir(Optmap.f_skeld, d, HXF_UID | HXF_GID | HXF_KEEP, uid, gid);
    }
    return 1;
}

int file_exists(const char *fn) {
    struct stat sb;
    return !stat(fn, &sb);
}

char *fixup_login_name(const char *ln, unsigned int idx, char *dest,
 size_t dsize) {
    /* Login names are fixup'ed (fixed) by adding a unique number them,
    necessarily truncating part of the original login name. */
    char idxstr[16], tln[16];
    snprintf(idxstr, 16, "%u", idx);
    snprintf(tln, 8 - strlen(idxstr) + 1, "%s", ln);
    snprintf(dest, dsize, "%s%s", tln, idxstr);
    return dest;
}

char *make_home(const char *ln, char *buf, size_t s) {
    if(Optmap.c_split == 1 || strlen(ln) == 1) {
        snprintf(buf, s, "%s/%c/%s", Optmap.f_home, *ln, ln);
    } else if(Optmap.c_split >= 2) {
        snprintf(buf, s, "%s/%c/%c%c/%s",
         Optmap.f_home, *ln, ln[0], ln[1], ln);
    } else {
        snprintf(buf, s, "%s/%s", Optmap.f_home, ln);
    }
    return buf;
}

char *make_login_name(const char *in_nname, const char *in_vname) {
    /* This takes the name (in_nname) and prenom (in_vname) and makes a 7-bit
    clean login name in the style VNNNNNN. Numerical indexes are done in
    fixup_login_name() (called from the main code). */
    char buf[8], fix_nname[8], fix_vname[8], *p;

    memset(fix_vname, '\0', 8);
    transform7(in_vname, fix_vname, 8);

    if(in_nname != NULL) {
        memset(fix_nname, '\0', 8);
        if((p = strchr(in_nname, ' ')) != NULL) { in_nname = ++p; }
        transform7(in_nname, fix_nname, 8);
        snprintf(buf, 8, "%c%s", *fix_vname, fix_nname);
    } else {
        snprintf(buf, 8, "%s", fix_vname);
    }

    if(strlen(buf) < 2) {
        // In case the login name is too short, for whatever reason, append a 0
        strcat(buf, "0");
    }
    return HX_strdup(HX_strlower(buf));
}

int only_digits(const char *p) {
    // Returns whether the string in P consists only of digits
    while(*p != '\0') {
        if(!isdigit(*p++)) { return 0; }
    }
    return 1;
}

char *read_file(const char *fn, const char *flags) {
    struct stat sb;
    char *dst;
    FILE *fp;

    if((fp = fopen(fn, flags)) == NULL || fstat(fileno(fp), &sb) != 0 ||
     (dst = malloc(sb.st_size + 1)) == NULL) {
        return NULL;
    }

    dst[sb.st_size - 1] = '\0';
    if(fread(dst, sb.st_size, 1, fp) < 1) {
        free(dst);
        return NULL;
    }

    fclose(fp);
    return dst;
}

//-----------------------------------------------------------------------------
static char *transform7(const byte *src, char *dest, size_t dsize) {
    /* Transform a name (possibly with umlauts, etc.) into a 7-bit clean,
    A-to-Z-only login name. */
    dest[--dsize] = '\0';
    while(dsize && *src != '\0') {
        if((*src >= 'A' && *src <= 'Z') || (*src >= 'a' && *src <= 'z')) {
            *dest++ = *src;
            --dsize;
        } else if(*src >= 128) {
            // This loop is still faster than a regex
            const struct stab *sp = subst_tab;
            while(sp->set != NULL) {
                // XXX: might want to use strpbrk() here
                if(strchr(sp->set, *src) != NULL) {
                    size_t l = strlen(sp->out);
                    if(dsize >= l) {
                        strncpy(dest, sp->out, l);
                        dest += l;
                        dsize -= l;
                    } else {
                        /* Do not let "abcdea" through, because if there was
                        too few space, the "" gets skipped but dsize is still
                        big enough to let the "a" in. */
                        dsize = 0;
                    }
                    break;
                }
                ++sp;
            }
        }
        ++src;
    }
    return dest;
}

static int wb_classof(char x) {
    if(isalpha(x)) { return CLASS_ALPHA; }
    if(isdigit(x)) { return CLASS_DIGIT; }
    if(x == '\0') { return CLASS_NONE; }
    return CLASS_OTHER;
}

//==[ End of file ]============================================================
