/*=============================================================================
Vitalnix/C-spark - User Management CLI/CUI
main.c - Startup code
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 1999 - 2004
  -- License restrictions apply (Dual FAL/GPL2)
  -- For details see doc/FAL.txt and doc/GPL2.txt.
  $Id: main.c,v 1.13 2004/05/11 14:05:55 hirogen2 Exp $
=============================================================================*/
#include <sys/file.h> /* lockf() */
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libHX.h>
#include <accdb.h>
#include "cui/global.h"
#include "cui/xml_in.h"

#include <locale.h>

enum {
    UG_ABRT   = 1 << 0,
    UG_PROMPT = 1 << 1,
    UG_DEFL   = 1 << 2,
};

static void print_banner(void);
static void init_var(int, char **);
static void init_applock(void);

static void ui_mainmenu(void);
static size_t ui_generic(const char *, char *, size_t,
 unsigned long opts, ...);

static void ui_sync(void);
static int ui_sync_get_log(char *, size_t);
static int ui_sync_prompt_cont_add(void);
static int ui_sync_prompt_cont_del(void);

static void ui_pwlist(void);
static int ui_pwlist_sty(char *, size_t);

static void ui_single(void);

static int count_args(const char **);
static int parse_options(int *, const char ***);
static int load_config(char *);
static int pconfig_aflush(const char *, unsigned char, void *, void *);
static int pconfig_crymeth(const char *, unsigned char, void *, void *);

struct optmap_s Optmap;
struct accdb_module *mm_output = NULL;

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
    int eax;
    init_var(argc, argv);
    init_applock();
    print_banner();

    if(!(eax = parse_options(&argc, (const char ***)&argv))) { return !eax; }
    if(!autorun_sd()) { ui_mainmenu(); }
    unlink(Optmap.f_lock);
    return EXIT_SUCCESS;
}

static void print_banner(void) {
    if(Optmap.nterm) {
        printf(
         "\e[1;31m"">" "\e[0;31m"">" "\e[1;30m""> "
         "\e[1;37m""V" "\e[0;37m""italnix" "\n\e[31m -- \e[0m"
        );
    } else {
        printf(">>> Vitalnix\n -- ");
    }

    printf(
      "C-spark %s <deemed stable> (%s)\n\n" // ]]
      "This program comes with ABSOLUTELY NO WARRANTY; it is free software,\n"
      "you are welcome to redistribute it under certain conditions;\n"
      "for details see the files doc/FAL.txt and doc/GPL2.txt which should\n"
      "have come with this program.\n\n",
      Optmap.progv, Optmap.progd
    );
    return;
}

static void init_var(int argc, char **argv) {
    const char *p;
    memset(&Optmap, 0, sizeof(Optmap));
    Optmap.progv = "v1.91.60.0";
    Optmap.progd = __DATE__ " " __TIME__;

    if((p = getenv("TERM")) != NULL && (strcmp(p, "linux") == 0 ||
     strcmp(p, "xterm") == 0 || strcmp(p, "msys") == 0 ||
     strcmp(p, "rxvt") == 0)) {
        ++Optmap.nterm;
    }

#ifdef __unix__
    if(!isatty(fileno(stdin)) && !isatty(fileno(stdout)) &&
     (p = getenv("DISPLAY")) != NULL && *p) {
        HX_fsystem(HX_FSYSTEM_ARGV | HX_FSYSTEM_EXEC, "xterm", argv,
         "-geometry", "80x25", "-bg", "#002040", "-fg", "gray", "-e", NULL);
        // exec() obviously failed when we get here
        fprintf(stderr, "HX_fsystem() unsuccessful: %s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }

#endif
    if(!isatty(fileno(stdout))) { Optmap.nterm = 0; }

    HX_strclone(&Optmap.f_home, "/home");
    HX_strclone(&Optmap.f_lock, "/var/run/cspark.pid");
    HX_strclone(&Optmap.f_shell, "/bin/bash");
    HX_strclone(&Optmap.f_skeld, "/var/lib/empty");
    Optmap.b_pwphon  = 1;
    Optmap.c_aflush  = AFLUSH_OFF;
    Optmap.c_crymeth = CRYPW_BLOWFISH;
    Optmap.c_split   = 0;
    Optmap.i_pswdlen = 10;

    umask(0077);
    load_config("/etc/spark.cfg");
    load_config("cui/../vetc/spark.cfg");
    load_config("spark.cfg");

    setlocale(LC_COLLATE, "");
    return;
}

static void init_applock(void) {
#ifndef _WIN32
    char buf[8];
    int fd;

    if((fd = open(Optmap.f_lock, O_RDWR | O_CREAT, 0666)) < 0) {
        perror("lockfile open()");
        exit(EXIT_FAILURE);
    }

    if(lockf(fd, F_TLOCK, 0) < 0) {
        perror("C-spark already running? lockf() returned");
        exit(EXIT_FAILURE);
    }

    ftruncate(fd, 0);
    snprintf(buf, 8, "%d\n", getpid());
    write(fd, buf, strlen(buf));
    fcntl(fd, F_SETFD, FD_CLOEXEC);
    // never close that file or the lock will be gone
#else
    return;
#endif
}

int Q_OPEN(void) {
    int eax;

    if((mm_output = accdb_load("shadow")) == NULL) {
        fprintf(stderr, "Could not load ACCDB back-end: %s\n",
         strerror(errno));
        return 0;
    }

    if((eax = mm_output->open(mm_output)) <= 0) {
        fprintf(stderr, "Could not open ACCDB back-end: %s"
         " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
        accdb_unload(mm_output);
        return 0;
    }

    return 1;
}

void Q_CLOSE(void) {
    if(mm_output != NULL) {
        mm_output->close(mm_output);
        accdb_unload(mm_output);
    }
    return;
}

//-----------------------------------------------------------------------------
static void ui_mainmenu(void) {
    byte m_running = 1;

    while(m_running) {
        char buf[MAXSNLEN];
        printf("\n"
          "MAIN MENU\n"
          "=========\n"
          "<2>  Synchronize user lists\n"
          "<4>  Print password list\n"
          "<5>  Add a single user\n"
          "<0>  Quit\n"
          "\n"
          "CHOICE > "
        );

        fflush(stdout);
        fgets(buf, MAXSNLEN, stdin);

        switch(*buf) {
            case '2':
                ui_sync();
                break;
            case '4':
                ui_pwlist();
                break;
            case '5':
                ui_single();
                break;
            case '9':
            case '0':
            case 'q':
                m_running = 0;
                break;
        }
    }

    return;
}

static size_t ui_generic(const char *msg, char *buf, size_t len,
 unsigned long opts, ...) {
    const char *defl = NULL;
    va_list argp;
    va_start(argp, opts);

    printf("\n");
    printf("%s", msg);
    if(opts & UG_ABRT) { printf("(Leave blank to abort.)\n"); }
    if(opts & UG_PROMPT) {
        const char *p = va_arg(argp, const char *);
        printf("%s ", p);
    }
    if(opts & UG_DEFL) {
        defl = va_arg(argp, const char *);
        printf("[%s] ", defl);
    }
    printf("> ");
    fflush(stdout);

    if(fgets(buf, len, stdin) == NULL) { return 0; }
    HX_chomp(buf);
    if(*buf == '\0' && opts & UG_DEFL) {
        buf[len - 1] = '\0';
        strncpy(buf, defl, len - 1);
    }

    return strlen(buf);
}

//-----------------------------------------------------------------------------
static void ui_sync(void) {
    struct HXbtree *addl = NULL, *lnlist = NULL;
    struct HXlld *dell = NULL;
    struct accdb_group dgroup;
    char buf[MAXFNLEN];
    FILE *logp = NULL;

    printf("\n\n"
      "IMPORT USERS\n"
      "============\n"
    );

    if(!Q_OPEN()) { return; }

    // ui_sync_get_dgrp()
    if(!ui_generic(
     "Enter the group to which compare users from the Data Source with.\n",
     buf, MAXFNLEN, UG_ABRT | UG_PROMPT, "Destination group")) {
        printf("No group specified, aborting.\n");
        goto out;
    }

    if(!sync_prep_dgrp(&dgroup, buf)) { goto out; }

    // ui_sync_get_file()
    if(!ui_generic(
     "Please enter the pathname to the file serving as input. The file must\n"
     "be in XML and must have the tags described in doc/sp_xmlds.html.\n",
     buf, MAXFNLEN, UG_ABRT | UG_PROMPT, "Input file")) {
        printf("No filename specified, aborting.\n");
        goto out;
    }

    if(sync_read_file(buf, &addl) <= 0) { goto out; }

    sync_compare(&dgroup, addl, &dell, &lnlist);
    if(addl->itemcount == 0 && dell->itemcount == 0) {
        printf("Already synchronized\n");
        goto out;
    }

    sync_fixup_addl(addl, lnlist);
    sync_set_aflush();
    if((logp = sync_open_log(buf, ui_sync_get_log(buf, MAXFNLEN))) == NULL) {
        goto out;
    }

    if(addl->itemcount > 0 && ui_sync_prompt_cont_add() &&
     !sync_add(&dgroup, addl, logp)) {
        printf("Error during operation. You might need to redo the"
         " Synchronization process\n" "to complete all jobs.\n");
        goto out2;
    }

    if(dell->itemcount > 0 && ui_sync_prompt_cont_del() &&
     !sync_del(dell, logp)) {
        printf("Error during operation. You might need to redo the"
         " Synchronization process\n" "to complete all jobs.\n");
    }

out2:
    /* We actually need to flush, since we have added home directores in /home,
    thus it would be nice to have at least some users have their account
    working. (Well, just keep a backup of your DB files, operations-so-far are
    recorded in the logfile and could thus be reversed easily.)
    Hm, Q_CLOSE also does an implicit flush if none has been done yet... */
    Q_SYSCTL(ACCDB_FLUSHDB);

out:
    if(logp != NULL) { fclose(logp); }
    sync_free(addl, dell);
    Q_CLOSE();
    return;
}

static int ui_sync_get_log(char *file, size_t len) {
    int eax = ui_generic(
      "Enter a name for a log file containing a preliminary list of added\n"
      "users and their respective plaintext passwords.\n",
      file, len, UG_ABRT | UG_PROMPT, "Log file"
    );

    if(eax == 0) { printf("No filename given, aborting.\n"); }
    if(eax <= 0) { return 0; }

    if(file_exists(file)) {
        char buf[MAXSNLEN];
        printf("\n"
          "File already exists. Specify what to do: [O]verwrite, "
          "[A]ppend or [C]ancel? (Leave blank to abort.)\n"
          "[Cancel] > "
        );

        fflush(stdout);
        fgets(buf, MAXSNLEN, stdin);
        HX_strlower(buf);
        if(*buf == 'a') { return 'a'; }
        else if(*buf == 'o') { return 'w'; }
        else if(*buf == '\0') {
            printf("No input, aborting.\n");
            return 0;
        } else {
            printf("You have to specify one of 'a', 'c' or 'o'. Aborting.\n");
            return 0;
        }
    }

    return 'w';
}

static int ui_sync_prompt_cont_add(void) {
    char buf[MAXSNLEN];
    printf("\n"
      "Continue with adding users? (Yes/No)\n"
      "[Yes] > "
    );

    fflush(stdout);
    fgets(buf, MAXSNLEN, stdin);
    HX_chomp(buf);
    HX_strlower(buf);

    if(*buf != '\0' && *buf != 'y') {
        printf("Skipping addition (interactive user request)\n");
        return 0;
    }

    return 1;
}

static int ui_sync_prompt_cont_del(void) {
    char buf[MAXSNLEN];
    printf("\n"
      "Continue with deleting users? (Yes/No)\n"
      "[Yes] > "
    );

    fflush(stdout);
    fgets(buf, MAXSNLEN, stdin);
    HX_chomp(buf);
    HX_strlower(buf);

    if(*buf != '\0' && *buf != 'y') {
        printf("Skipping deletion (interactive user request)\n");
        return 0;
    }

    return 1;
}

//-----------------------------------------------------------------------------
static void ui_pwlist(void) {
    char infile[MAXFNLEN], outfile[MAXFNLEN], style[MAXSNLEN];

    printf("\n\n"
     "PRINT PASSWORD LISTS\n"
     "====================\n"
    );

    if(!ui_generic("Enter the name of the logfile that was created during"
     " user list\n" "synchronization.\n", infile, MAXFNLEN,
     UG_ABRT | UG_PROMPT, "Input log file")) {
        printf("No filename specified, aborting.\n");
        return;
    }

    if(!ui_generic("Enter the name of the file where the formatted"
     " (\"beautified\")\n" "output should go.\n", outfile, MAXFNLEN,
     UG_ABRT | UG_PROMPT, "Output file")) {
        printf("No filename specified, aborting.\n");
        return;
    }

    if(!ui_pwlist_sty(style, MAXFNLEN)) { return; }
    pwl_proc(style, infile, outfile);
    return;
}

static int ui_pwlist_sty(char *buf, size_t len) {
    printf("\n"
      "Enter the canoncial name of a function responsible for output\n"
      "formatting. Enter \"?\" (without quotes) to see a list of names.\n"
      "(Leave blank to abort.)\n"
    );

    while(1) {
        printf("Style > ");
        fflush(stdout);
        fgets(buf, len, stdin);
        HX_chomp(buf);
        if(*buf == '?') {
            pwl_styles();
            continue;
        } else if(*buf == '\0') {
            return 0;
        }
        break;
    }

    return 1;
}

//-----------------------------------------------------------------------------
static void ui_single(void) {
    static const char *msg[] = {
        "Enter the last name of the user to be added.\n", "Last name",
        "Enter the first name of the user to be added.\n", "First name",
        "Enter the XUID of the new user (may be empty)\n", "XUID",
        "Enter the group name or GID to add the user to.\n", "Group",
        "Enter the subgroup of the new user (may be empty)\n", "SGRP",
    };
    char nname[MAXSNLEN], vname[MAXSNLEN], xuid[MAXSNLEN], group[MAXSNLEN],
     sgroup[MAXSNLEN];

    printf("\n\n"
     "ADD A SINGLE USER\n"
     "=================\n"
    );

    ui_generic(msg[0], nname,  MAXSNLEN, UG_ABRT | UG_PROMPT, msg[1]);
    ui_generic(msg[2], vname,  MAXSNLEN, UG_ABRT | UG_PROMPT, msg[3]);
    ui_generic(msg[4], xuid,   MAXSNLEN, UG_ABRT | UG_PROMPT, msg[6]);
    ui_generic(msg[6], group,  MAXSNLEN, UG_ABRT | UG_PROMPT, msg[8]);
    ui_generic(msg[8], sgroup, MAXSNLEN, UG_ABRT | UG_PROMPT, msg[9]);
    return;
}

//-----------------------------------------------------------------------------
static int count_args(const char **args) {
    int argk = 0;
    while(*args++ != NULL) { ++argk; }
    return argk;
}

static int load_config(char *fn) {
    static struct rconfig_opt options_table[] = {
        {"AUTOFLUSH",      RCONF_CB,     NULL,                   pconfig_aflush},
        {"HOME",           RCONF_STRING, &Optmap.f_home,         NULL},
        {"MASTER_PREADD",  RCONF_STRING, &Optmap.master_preadd,  NULL},
        {"MASTER_PREDEL",  RCONF_STRING, &Optmap.master_predel,  NULL},
        {"MASTER_POSTADD", RCONF_STRING, &Optmap.master_postadd, NULL},
        {"MASTER_POSTDEL", RCONF_STRING, &Optmap.master_postdel, NULL},
        {"PSWD_LEN",       RCONF_INT,    &Optmap.i_pswdlen,      NULL},
        {"PSWD_METH",      RCONF_CB,     NULL,                   pconfig_crymeth},
        {"PSWD_PHON",      RCONF_BOOL,   &Optmap.b_pwphon,       NULL},
        {"SHELL",          RCONF_STRING, &Optmap.f_shell,        NULL},
        {"SKEL",           RCONF_STRING, &Optmap.f_skeld,        NULL},
        {"SPLIT_LVL",      RCONF_UCHAR,  &Optmap.c_split,        NULL},
        {"USER_PREADD",    RCONF_STRING, &Optmap.user_preadd,    NULL},
        {"USER_PREDEL",    RCONF_STRING, &Optmap.user_predel,    NULL},
        {"USER_POSTADD",   RCONF_STRING, &Optmap.user_postadd,   NULL},
        {"USER_POSTDEL",   RCONF_STRING, &Optmap.user_postdel,   NULL},
        {NULL},
    };

    return accdb_rconfig(fn, options_table);
}

static int parse_options(int *argc, const char ***argv) {
    static const char *_empty_argv[] = {NULL};
    char *tmp;
    struct poptOption options_table[] = {
        {NULL, 'P', POPT_ARG_STRING, &tmp, 'P',
         "Print password list", "style:infile:outfile"},
        {NULL, 'S', POPT_ARG_STRING, &tmp, 'S',
         "Synchronize with list", "group:infile[:logfile]"},
        {NULL, 'U', POPT_ARG_STRING, &tmp, 'U',
         "Add single user", "nname:vname:bday:sgroup"},
        {NULL, 'c', POPT_ARG_STRING, &tmp, 'c',
         "Load alternate config file", "file"},
        {NULL, 'h', POPT_ARG_NONE, NULL, 'h', "Show this help message", NULL},
        {"version", 'v', POPT_ARG_NONE, NULL, 'v',
         "Show version number and exit", NULL},

        {"template", '\0', POPT_ARG_STRING, &Optmap.pwl_template, '\0',
         "A file template for some PWL styles", "FILE"},
        POPT_AUTOHELP
        POPT_TABLEEND
    };

    poptContext ctx;
    const char **args;
    int c, argk;

    ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
    while((c = poptGetNextOpt(ctx)) >= 0) {
        switch(c) {
            case 'P':
                autorun_pwl(tmp);
                free(tmp);
                break;
            case 'S':
                autorun_sync(tmp);
                free(tmp);
                break;
            case 'U':
                autorun_user(tmp);
                free(tmp);
                break;
            case 'c':
                if(!load_config(tmp)) {
                     fprintf(stderr, "Parsing of %s unsuccessful\n", tmp);
                }
                free(tmp);
                break;
            case 'h':
                poptPrintHelp(ctx, stdout, 0);
                exit(EXIT_SUCCESS);
                break;
            case 'v':
                printf("C-spark %s (%s)\n", Optmap.progv, Optmap.progd);
                exit(EXIT_SUCCESS);
                break;
        }
    }

    if(c < -1) {
        fprintf(stderr, "cspark: %s: %s\n",
         poptBadOption(ctx, 0), poptStrerror(c));
        poptFreeContext(ctx);
        return 0;
    }

    if((args = poptGetArgs(ctx)) != NULL) {
        argk = count_args(args);
        poptDupArgv(argk, args, argc, argv);
    } else {
        *argc = 0;
        *argv = _empty_argv;
    }

    return 1;
}

static int pconfig_aflush(const char *key, unsigned char type, void *ptr,
 void *uptr) {
    unsigned char *v = &Optmap.c_aflush;
    if(stricmp(ptr, "off") == 0)          { *v = AFLUSH_OFF;     }
    else if(stricmp(ptr, "on") == 0)      { *v = AFLUSH_ON;      }
    else if(stricmp(ptr, "postadd") == 0) { *v = AFLUSH_POSTADD; }
    else { *v = AFLUSH_DEFAULT; }
    return 1;
}

static int pconfig_crymeth(const char *key, unsigned char type, void *ptr,
 void *uptr) {
    unsigned char *v = &Optmap.c_crymeth;
    if(stricmp(ptr, "md5") == 0)      { *v = CRYPW_MD5; }
    else if(stricmp(ptr, "des") == 0) { *v = CRYPW_DES; }
    else { *v = CRYPW_BLOWFISH; }
    return 1;
}

//==[ End of file ]============================================================
