/*=============================================================================
Vitalnix/C-spark - User Management CLI/CUI
pwl.c
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 1999 - 2004
  -- License restrictions apply (Dual FAL/GPL2)
  -- For details see doc/FAL.txt and doc/GPL2.txt.
  $Id: pwl.c,v 1.1 2004/05/11 13:34:35 hirogen2 Exp $
=============================================================================*/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libHX.h>
#include "cui/global.h"

#define PROT_RDTMPL struct style *S
#define PROT_HEADER struct style *S, FILE *fp, const char *file
#define PROT_THDR   struct style *S, FILE *fp, const char *sg
#define PROT_ENTRY  struct style *S, FILE *fp, const char *sg, \
  const char *vname, const char *nname, const char *user, const char *pw
#define PROT_TFOOT  struct style *S, FILE *fp
#define PROT_FOOTER struct style *S, FILE *fp

enum {
    TPAR_HEADER = 0,
    TPAR_THDR,
    TPAR_ENTRY,
    TPAR_TFOOT,
    TPAR_FOOTER,
    TPAR_LASTMARK,
};

struct style {
    const char *name, *desc;
    void (*rd_tmpl)(PROT_RDTMPL);
    void (*header)(PROT_HEADER);
    void (*thdr)(PROT_THDR);
    void (*entry)(PROT_ENTRY);
    void (*tfoot)(PROT_TFOOT);
    void (*footer)(PROT_FOOTER);
    char *tp, *tpp[TPAR_LASTMARK];
};

// Function prototypes
static struct HXbtree *build_tree(FILE *);
static int pwl_proc2(struct style *, FILE *, const char *);
static int pwl_proc3(struct style *, struct HXbtree *, FILE *, const char *);

// Prototypes for styles
static void sg_html_header(PROT_HEADER);
static void sg_html_thdr(PROT_THDR);
static void sg_html_entry(PROT_ENTRY);
static void sg_html_tfoot(PROT_TFOOT);
static void sg_html_footer(PROT_TFOOT);

static void sg_rtf_rdtmpl(PROT_RDTMPL);
static void sg_rtf_header(PROT_HEADER);
static void sg_rtf_thdr(PROT_THDR);
static void sg_rtf_entry(PROT_ENTRY);
static void sg_rtf_tfoot(PROT_TFOOT);
static void sg_rtf_footer(PROT_TFOOT);

static void sg_txt_thdr(PROT_THDR);
static void sg_txt_entry(PROT_ENTRY);
static void sg_txt_tfoot(PROT_TFOOT);

static void a1_header(PROT_HEADER);
static void a1_entry(PROT_ENTRY);

static void sb_rdtmpl(PROT_RDTMPL);
static void sb_header(PROT_HEADER);
static void sb_entry(PROT_ENTRY);
static void sb_footer(PROT_FOOTER);

// Variables
static struct style styles_table[] = {
    /* "sg": Per-sgroup sorted overview.
    Each subgroup (whose members will be sorted by name) gets its own table.
    Useful for the annual grade shift. */
    {name: "sg_html", desc: "Per-sgroup tables (HTML)", rd_tmpl: NULL,
    header: sg_html_header, thdr: sg_html_thdr, entry: sg_html_entry,
    tfoot: sg_html_tfoot, footer: sg_html_footer},

    {name: "sg_rtf", desc: "Per-sgroup tables (RTF)", rd_tmpl: sg_rtf_rdtmpl,
    header: sg_rtf_header, thdr: sg_rtf_thdr, entry: sg_rtf_entry,
    tfoot: sg_rtf_tfoot, footer: sg_rtf_footer},

    {name: "sg_txt", desc: "Per-sgroup tables (TXT)", rd_tmpl: NULL,
    header: NULL, thdr: sg_txt_thdr, entry: sg_txt_entry, tfoot: sg_txt_tfoot,
    footer: NULL},

    /* "a1": All-in-1 sorted overview.
    Input data is sorted by sgroup first, then name and is presented in one
    table. Useful for small user changes or administration-only. (Thus, a
    text/plain version will suffice. */
    {name: "a1", desc: "Single sorted view (TXT)", rd_tmpl: NULL,
    header: a1_header, thdr: NULL, entry: a1_entry, tfoot: sg_txt_tfoot,
    footer: NULL},

    /* "sb": Serienbrief (letter series)
    Originally brought into Dautool/Daxtraq in 2002 by Cristoph Thiel */
    {name: "sb", desc: "Serienbrief (letter series) (LATEX)",
    rd_tmpl: sb_rdtmpl, header: sb_header, thdr: NULL, entry: sb_entry,
    tfoot: NULL, footer: sb_footer},

    {NULL},
};

//-----------------------------------------------------------------------------
int pwl_proc(const char *style, const char *infile, const char *outfile) {
    // General loader for style processing. Not optimal.
    struct style *sp = styles_table;
    FILE *ifp;

    if(style == NULL) { return 0; }
    while(sp->name != NULL) {
        if(strcmp(style, sp->name) == 0) { break; }
        ++sp;
    }

    if(sp->name == NULL) {
        fprintf(stderr, "No such style: \"%s\"\n", style);
        return 0;
    }

    if((ifp = fopen(infile, "r")) == NULL) {
        fprintf(stderr, "Could not open %s: %s\n", infile, strerror(errno));
        return 0;
    }

    printf("Processing \"%s\" to \"%s\" using \"%s\"\n",
     infile, outfile, style);
    return pwl_proc2(sp, ifp, outfile);
}

void pwl_styles(void) {
    struct style *sp = styles_table;
    printf("Available styles:\n");
    while(sp->name != NULL) {
        printf("\t%s:\t%s\n", sp->name, sp->desc);
        ++sp;
    }
    return;
}

//-----------------------------------------------------------------------------
static int pwl_proc2(struct style *style, FILE *ip, const char *outf) {
    struct HXbtree *bt;
    FILE *op;
    int rv = 1;

    if(style->rd_tmpl != NULL && Optmap.pwl_template == NULL) {
        fprintf(stderr, "pwl_proc2(): You will need to specify a template for"
         " this style.\n" "Use --template=FILE (specify it before -P).\n");
        return 0;
    }

    if((bt = build_tree(ip)) == NULL) {
        fprintf(stderr, "build_tree(): %s\n", strerror(errno));
        fclose(ip);
        return 0;
    }

    fclose(ip);
    if((op = fopen(outf, "w")) == NULL) {
        fprintf(stderr, "Could not open %s: %s\n", outf, strerror(errno));
        return 0;
    }

    rv = pwl_proc3(style, bt, op, outf);
    fclose(op);
    HXbtree_free(bt);
    printf("\n");
    return rv;
}

static int pwl_proc3(struct style *S, struct HXbtree *bt, FILE *ofp,
 const char *file) {
    /* In 99% of all cases, administrators want (any kind of) output to be
    sorted, even if it is just some small number of users. Thus, the
    implemented "architecture" for realization includes callbacks for
    pwl_proc3(). */
    struct HXbtrav *travp;
    struct HXbtree_node *nd;

    if(S->rd_tmpl != NULL) { S->rd_tmpl(S); }
    if(S->header != NULL) { S->header(S, ofp, file); }
    if((travp = HXbtrav_init(bt, NULL)) == NULL) {
        fprintf(stderr, "pwl_proc3(): Initializing traverser failed: %s\n",
         strerror(errno));
        return 0;
    }

    while((nd = HXbtraverse(travp)) != NULL) {
        struct HXbtrav *sg = HXbtrav_init(nd->data, NULL);
        struct HXbtree_node *sd;
        if(S->thdr != NULL) { S->thdr(S, ofp, nd->key); }

        while((sd = HXbtraverse(sg)) != NULL) {
            char *ptr = sd->key, *c_nname, *c_vname, *c_user, *c_pw;
            c_nname = HX_strsep(&ptr, ":");
            c_vname = HX_strsep(&ptr, ":");
            c_user  = HX_strsep(&ptr, ":");
            c_pw    = HX_strsep(&ptr, ":");
            if(S->entry != NULL) {
                S->entry(S, ofp, nd->key, c_nname, c_vname, c_user, c_pw);
            }
        }

        if(S->tfoot != NULL) { S->tfoot(S, ofp); }
        HXbtree_free(nd->data);
    }

    if(S->footer != NULL) { S->footer(S, ofp); }
    HXbtrav_free(travp);
    return 1;
}

static struct HXbtree *build_tree(FILE *ifp) {
    /* This function reads a logfile as generated by the synchronization stage,
    puts it in a B-tree (again) just to retrieve it sorted afterwards. */
    struct HXbtree *bt = HXbtree_init(HXBT_ASSOC | HXBT_PCMP, compare_wbc);
    char buf[MAXLNLEN];
    int std = 0;

    if(bt == NULL) { return NULL; }

    while(fgets(buf, MAXLNLEN, ifp) != NULL) {
        char *bufp = buf, *c_type, *c_sgroup;
        struct HXbtree *sgp;

        HX_chomp(buf);
        if(*buf == '\0') { continue; }
        if(*buf == '#') {
            if(!std && strncmp(buf, "# $logformat", 12) == 0 &&
             strcmp(buf + 12, " " LOGFMT_VERSION) != 0) {
                fprintf(stderr, "Input file is not of version \""
                 LOGFMT_VERSION "\"\n");
                HXbtree_free(bt);
                return 0;
            }
            continue;
        }

        /* Won't take any $logformat into account now since data has
        started. */
        ++std;

        c_type = HX_strsep(&bufp, ":");
        c_sgroup = HX_strsep(&bufp, ":");
        if(*c_type != '+' || c_sgroup == NULL || bufp == NULL) { continue; }

        // Only print users which were added ('+')
        if((sgp = HXbtree_get(bt, c_sgroup)) == NULL) {
            sgp = HXbtree_init(HXBT_CPTR | HXBT_PCMP, strcoll);
            HXbtree_add(bt, c_sgroup, sgp);
        }
        HXbtree_add(sgp, bufp);
    }
    return bt;
}

//-----------------------------------------------------------------------------
/* Please do not remove the "Formatted by" messages as they are the only
external advertisements for this software. */

static void sg_html_header(PROT_HEADER) {
    fprintf(fp, "<html><head><title>%s</title><style type=\"text/css\">\n"
     "body { font: 12pt \"Arial\", \"Sans Serif\"; }\n"
     "th { text-align: left; }\n"
     "</style><body>", file);
    return;
}

static void sg_html_thdr(PROT_THDR) {
    fprintf(fp, "<table width=\"100%%\"><tr><td><h1>Group %1$s</h1></td>"
     "<td align=\"right\"><i>Formatted by Vitalnix C-Spark %2$s</i></td></tr>"
     "</table><table cellpadding=\"5\" width=\"100%%\"><tr>"
     "<th width=\"50%%\">Name</th><th>Login</th><th>Password</th></tr>\n",
     sg, Optmap.progv);
    return;
}

static void sg_html_entry(PROT_ENTRY) {
    char buf[MAXSNLEN];
    snprintf(buf, MAXSNLEN, "%s, %s", nname, vname);
    fprintf(fp, "<tr><td>%s</td><td>%s</td><td>%s</td></tr>\n",
     buf, user, pw);
    return;
}

static void sg_html_tfoot(PROT_TFOOT) {
    fprintf(fp, "</table>");
    return;
}

static void sg_html_footer(PROT_FOOTER) {
    fprintf(fp, "</body></html>\n");
    return;
}

//-----------------------------------------------------------------------------
static void sg_rtf_rdtmpl(PROT_RDTMPL) {
    char *p;
    if((S->tp = read_file(Optmap.pwl_template, "r")) == NULL) {
        fprintf(stderr, "sg_rtf_rdtmpl(): Could not open template file"
         " %s: %s\n", Optmap.pwl_template, strerror(errno));
        exit(EXIT_FAILURE);
    }

    /* The following fixup is so that the RTF file can be edited more easily.
    And of course, we keep it in one file, which is essential. */
    S->tpp[TPAR_HEADER] = p = S->tp;

    /* Micro$oft Office successfully ignores our private tags, so that we can
    check out what our template looks like without having to wade through. */

#define SEP(__str, __len, __ent) /* some small strstr_strsep() */ \
     p = strstr(p, (__str)); *p = '\0'; p += (__len); S->tpp[__ent] = p;
    SEP("{\\THDR}",   7, TPAR_THDR);
    SEP("{\\ENTRY}",  8, TPAR_ENTRY);
    SEP("{\\TFOOT}",  8, TPAR_TFOOT);
    SEP("{\\FOOTER}", 9, TPAR_FOOTER);
#undef SEP
    return;
}

static void sg_rtf_header(PROT_HEADER) {
    fprintf(fp, S->tpp[TPAR_HEADER], Optmap.progv, Optmap.progv, file);
    return;
}

static void sg_rtf_thdr(PROT_THDR) {
    fprintf(fp, S->tpp[TPAR_THDR], sg);
    return;
}

static void sg_rtf_entry(PROT_ENTRY) {
    fprintf(fp, S->tpp[TPAR_ENTRY], sg, nname, vname, user, pw);
    return;
}

static void sg_rtf_tfoot(PROT_TFOOT) {
    fprintf(fp, S->tpp[TPAR_TFOOT]);
    return;
}

static void sg_rtf_footer(PROT_FOOTER) {
    fprintf(fp, S->tpp[TPAR_FOOTER]);
    return;
}

//-----------------------------------------------------------------------------
static void sg_txt_thdr(PROT_THDR) {
    fprintf(fp, "\n\n"
     "                                  Formatted by Vitalnix C-Spark %s\n\n\n"
     "  >>> Group %s\n\n"
     "  %-35s  %-10s  %s\n"
     "  ====================================="
     "==================================\n",
     Optmap.progv, sg, "Name", "Login", "Password");
    return;
}

static void sg_txt_entry(PROT_ENTRY) {
    char buf[MAXSNLEN];
    snprintf(buf, MAXSNLEN, "%s, %s", nname, vname);
    fprintf(fp, "  %-35s  %-10s  %s\n", buf, user, pw);
    return;
}

static void sg_txt_tfoot(PROT_TFOOT) {
    fprintf(fp, "\x0C");
    return;
}

//-----------------------------------------------------------------------------
static void a1_header(PROT_HEADER) {
    fprintf(fp, "\n\n"
     "                                  Formatted by Vitalnix C-Spark %s\n\n\n"
     "  %-8s  %-35s  %-10s  %s\n"
     "  ====================================="
     "==================================\n",
     Optmap.progv, "Group", "Name", "Login", "Password");
    return;
}

static void a1_entry(PROT_ENTRY) {
    char buf[MAXSNLEN];
    snprintf(buf, MAXSNLEN, "%s, %s", nname, vname);
    fprintf(fp, "  %-8s  %-35s  %-10s  %s\n", sg, buf, user, pw);
    return;
}

//-----------------------------------------------------------------------------
static void sb_rdtmpl(PROT_RDTMPL) {
    /* Unlike sg_rtf, this template should only contain the
    THDR/ENTRY parts! (TFOOT is handled in sb_entry().) */
    if((S->tp = read_file(Optmap.pwl_template, "r")) == NULL) {
        fprintf(stderr, "sb_rdtmpl(): Could not open template file %s: %s\n",
         Optmap.pwl_template, strerror(errno));
        exit(EXIT_FAILURE);
    }
    return;
}

static void sb_header(PROT_HEADER) {
    fprintf(fp,
      "\\documentclass[12pt]{article}\n\n"
      "\\usepackage{isolatin1}\n"
      "\\usepackage{graphics}\n"
      "\\usepackage{german}\n\n"
      "\\begin{document}\n\n"
    );
    return;
}

static void sb_entry(PROT_ENTRY) {
    fprintf(fp, S->tp, sg, nname, vname, user, pw);
    fprintf(fp, "\\newpage\n");
    return;
}

static void sb_footer(PROT_FOOTER) {
    fprintf(fp, "\\end{document}\n");
    free(S->tp);
    return;
}

//==[ End of file ]============================================================
