/*=============================================================================
Vitalnix/C-spark - User Management CLI/CUI
sync.c - Where the real work happens
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 1999 - 2004
  -- License restrictions apply (Dual FAL/GPL2)
  -- For details see doc/FAL.txt and doc/GPL2.txt.
  $Id: sync.c,v 1.3 2004/05/11 14:01:38 hirogen2 Exp $
=============================================================================*/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <libHX.h>
#include <accdb.h>
#include "cui/global.h"
#include "cui/ds_entry.h"
#include "cui/xml_in.h"

inline static void free_addl_entry(struct ds_entry *);
inline static void handle_user(struct accdb_user *, struct HXbtree *,
 struct HXlld *, size_t *);

//-----------------------------------------------------------------------------
int sync_prep_dgrp(struct accdb_group *grp, const char *name) {
    /* Here, we request a group to which compare the data source against.
    Usually, a group is always the same for a particular input set otherwise
    strange things may happen. You can either enter the group name or a GID.
    (Once there is a non-digit in your input, it is treated as a group
    name.) */
    int eax;

    if(only_digits(name)) {
        grp->gid = strtol(name, NULL, 0);
        grp->gname = NULL;
    } else {
        grp->gid = -1;
        grp->gname = (char *)name;
    }

    if((eax = Q_GROUPINFO(grp, grp, 1)) < 0) {
        fprintf(stderr, "Error querying the ACCDB: %s (eax=%d, errno=%d)\n",
         strerror(errno), eax, errno);
        return 0;
    } else if(eax == 0) {
        fprintf(stderr, "Group \"%s\" does not exist\n", name);
        return 0;
    }

    /* After [successful] groupinfo(), dgroup->gname is now a pointer to space
    malloc'ed from within the backend. So, dgroup->gname can safely be used
    after this line. */
    return 1;
}

int sync_read_file(const char *file, struct HXbtree **addl) {
    /* In this part, we stuff all users from the data source into an
    associative array, called addl. Some elements of that array will further be
    changed until it is ready for running through import_ds(). */
    struct ds_entry *bucket;
    char buf[MAXLNLEN];
    void *state;
    int eax;

    if((eax = xmlds_open(file, &state)) <= 0) { return eax; }

    /* This loop fills the addl associative array by generating a suited
    (pre-stage) login name. (These login names may already be used by the
    system and must first be fixed up (i.e. by adding a unique postcount
    number), so that there are only unique login name in the user database.) */

    *addl = HXbtree_init(HXBT_ASSOC);

    while(1) {
        bucket = malloc(sizeof(struct ds_entry));

        if((eax = xmlds_read(state, bucket)) < 0) {
            fprintf(stderr, "Warning: Stumbled upon a bogus entry"
             " in Data Source, continuing\n");
            free(bucket);
            continue;
        } else if(eax == 0) {
            // End of list
            free(bucket);
            break;
        }

        bucket->lname = make_login_name(bucket->nname, bucket->vname);

        if(bucket->nname != NULL) {
            snprintf(buf, MAXLNLEN, "%s %s", bucket->vname, bucket->nname);
            bucket->rname = HX_strdup(buf);
        } else {
            bucket->rname = HX_strdup(bucket->vname);
        }

        if(bucket->xuid != NULL) {
            snprintf(buf, MAXLNLEN, "%s,%s", bucket->rname, bucket->xuid);
        } else {
            buf[MAXLNLEN - 1] = '\0';
            strncpy(buf, bucket->rname, MAXLNLEN - 1);
        }
        HXbtree_add(*addl, buf, bucket);
    }

    xmlds_close(state);
    return 1;
}

void sync_compare(struct accdb_group *dgroup, struct HXbtree *addl,
 struct HXlld **dell, struct HXbtree **lnlist) {
    /* And this is now the comparison algorithm, (half of) the core. While
    traversing the users, we mark used login names by putting them in an array
    (lnlist). Login names of users which are about to be deleted are also
    pushed there, so they do not get reassigned right now, but at the next
    run-through. */
    size_t users_grp = 0, users_keep = 0;
    struct accdb_user passwd;

    printf("\n" "Comparing Data Source against UDB\n");
    *lnlist = HXbtree_init(HXBT_CPTR | HXBT_PCMP, strcmp);
    *dell = HXlld_init(0);

    while(Q_USERTRAV(&passwd)) {
        HXbtree_add(*lnlist, passwd.lname);

        if((dgroup->gid == -1 || dgroup->gid == passwd.gid) &&
         (dgroup->gname == NULL || passwd.igrp == NULL ||
         strcmp(dgroup->gname, passwd.igrp) == 0)) {
            // Users we look for (only specified group)
            ++users_grp;
            handle_user(&passwd, addl, *dell, &users_keep);
        }
    }

    printf("%u group member%s found\n", (unsigned int)users_grp,
     users_grp != 1 ? "s" : "");
    printf("   %u to keep\n", (unsigned int)users_keep);
    printf("   %lu to delete\n", (**dell).itemcount);
    printf("%lu to add\n", addl->itemcount);
    return;
}

void sync_fixup_addl(struct HXbtree *addl, struct HXbtree *lnlist) {
    // Here comes the fixup loop (other part of the core).
    // DEVNOTES: frees "lnlist"
    void *travp = HXbtrav_init(addl, NULL);
    struct HXbtree_node *node;
    char sbuf[16];

    if(addl->itemcount == 0) { return; }

    printf("Username fixup...\n");
    while((node = HXbtraverse(travp)) != NULL) {
        struct ds_entry *entry = node->data;
        int i = 0;
        sbuf[15] = '\0';
        strncpy(sbuf, entry->lname, 15);

        /* If the fixup'ed login name with the unique number already exists,
        bump the unique number. */
        while(HXbtree_find(lnlist, sbuf) != NULL) {
            fixup_login_name(entry->lname, ++i, sbuf, 16);
        }

        if(strcmp(entry->lname, sbuf) != 0) {
            HX_strclone(&entry->lname, sbuf);
        }
        HXbtree_add(lnlist, sbuf);
    }

    HXbtrav_free(travp);
    HXbtree_free(lnlist);
    return;
}

void sync_set_aflush(void) {
    if(Optmap.c_aflush == AFLUSH_ON) {
        Q_SYSCTL(ACCDB_ADDFLAGS, ACCDB_AFLUSH);
    } else if(Optmap.c_aflush == AFLUSH_DEFAULT) {
        unsigned long f = Q_SYSCTL(ACCDB_ADDFLAGS, 0);
        printf("Notice: AUTOFLUSH is turned %s\n",
         (f & ACCDB_AFLUSH) ? "on" : "off");
    } else {
        Q_SYSCTL(ACCDB_DELFLAGS, ACCDB_AFLUSH);
    }
    return;
}

FILE *sync_open_log(const char *file, char mc) {
    char mode[4] = {mc, '\0'};
    FILE *rv;

    if(mc == '\0') { return NULL; }
    if((rv = fopen(file, mode)) == NULL) {
        fprintf(stderr, "Error: Could not open %s: %s\n",
         file, strerror(errno));
        return NULL;
    }

    // change the number if the format should ever change
    fprintf(rv, "# $logformat 0\n");

    fprintf(rv, "# C-Spark %s\n", Optmap.progv);
    return rv;
}

int sync_add(struct accdb_group *dgroup, struct HXbtree *addl, FILE *logp) {
    // Adding the users to the system database
    ssize_t ic = addl->itemcount;
    struct HXbtree_node *node;
    char *plain_pw;
    void *travp;
    int eax;

    if(ic == 0) { return 1; } // Nothing to do

    if(Optmap.master_preadd != NULL) {
        char cmd[MAXLNLEN];
        snprintf(cmd, MAXLNLEN, Optmap.master_preadd, (unsigned int)ic);
        system(cmd);
    }

    plain_pw = malloc(Optmap.i_pswdlen + 1);
    travp = HXbtrav_init(addl, NULL);
    while((node = HXbtraverse(travp)) != NULL) {
        struct ds_entry *in = node->data;
        struct accdb_user out;
        char uhome[MAXFNLEN];

        memset(&out, 0, sizeof(struct accdb_user));
        out.lname = in->lname;
        out.uid   = -1;
        out.gid   = dgroup->gid;
        out.gecos = node->key;
        out.home  = make_home(in->lname, uhome, MAXFNLEN);
        out.shell = Optmap.f_shell;
        out.igrp  = dgroup->gname;
        out.sgrp  = NULL;

        if(Optmap.i_pswdlen < 0) {
            out.passwd = "!";
            out.last_change = 0;
        } else if(Optmap.i_pswdlen == 0) {
            out.passwd = "";
            out.last_change = (unsigned long)time(NULL) / 86400; // "now"
        } else {
            unsigned long fl = GENPW_ONE_DIGIT | GENPW_ONE_CASE;
            if(Optmap.b_pwphon) { fl |= GENPW_PHONEMIC; }
            accdb_genpw(plain_pw, Optmap.i_pswdlen, fl);
            accdb_cryptpw(plain_pw, NULL, Optmap.c_crymeth, &out.passwd);
            out.last_change = (unsigned long)time(NULL) / 86400;
        }

        out.keep_min    = 0;
        out.keep_max    = 366;
        out.warn_age    = 21;
        out.expire      = -1;
        out.inactive    = -1;

        if(Optmap.user_preadd != NULL) {
            /* Order of positional parameters is important, as we should stay
            compatible in future. See vetc/spark.cfg or doc/spark_ui.html for
            the assigned order. */
            char cmd[MAXLNLEN];
            snprintf(cmd, MAXLNLEN, Optmap.user_preadd,
             out.lname, out.uid, out.gid, out.igrp, out.sgrp,
             out.gecos, out.home, out.shell);
            system(cmd);
        }

        if((eax = Q_USERADD(&out)) <= 0) {
            memset(plain_pw, '\0', Optmap.i_pswdlen + 1);
            free(plain_pw);
            printf("\n");
            fprintf(stderr, "useradd: Something went wrong. (eax=%d,"
             " errno=%d: %s)\n", eax, errno, strerror(errno));
            HXbtrav_free(travp);
            return 0;
        }

        if(logp != NULL) {
            fprintf(logp, "+:%s:%s:%s:%s:%s\n", in->sgroup, in->nname,
             in->vname, out.lname, plain_pw);
        }

        if(Optmap.i_pswdlen > 0) { free(out.passwd); }

        // Create home directory and optionally copy skeleton dir
        if(!create_home(out.home, out.uid, out.gid)) {
            printf("\n");
            fprintf(stderr, "Warning: Could not create home directory"
             "%s: %s\n", out.home, strerror(errno));
        }

        if(Optmap.user_postadd != NULL) {
            char cmd[MAXLNLEN];
            snprintf(cmd, MAXLNLEN, Optmap.user_preadd,
             out.lname, out.uid, out.gid, out.igrp, out.sgrp,
             out.gecos, out.home, out.shell);
            if(Optmap.c_aflush == AFLUSH_POSTADD) { Q_SYSCTL(ACCDB_FLUSHDB); }
            system(cmd);
        }

        if((--ic & 0x3F) == 0) {
            printf("add: %u users remaining\n", (unsigned int)ic);
        }
    }

    printf("Successfully added %u users\n", (unsigned int)addl->itemcount);
    memset(plain_pw, '\0', Optmap.i_pswdlen + 1);
    free(plain_pw);
    HXbtrav_free(travp);

    if(Optmap.master_postadd != NULL) {
        char cmd[MAXLNLEN];
        snprintf(cmd, MAXLNLEN, Optmap.master_postadd,
         (unsigned int)addl->itemcount);
        system(cmd);
    }

    return 1;
}

int sync_del(struct HXlld *dell, FILE *logp) {
    // Deleting the old users
    size_t ic = dell->itemcount;
    struct HXlld_node *travp;
    char buf[MAXSNLEN];
    int eax;

    if(ic == 0) { return 1; } // Nothing to do

    /* We generate a time stamp here so that a possible backup operation will
    write into the same directory for all users to be deleted. The format
    should also stay the same (compatibility reasons), if possible. */
    {
#ifdef _WIN32
#    define localtime_r(__now, __res) localtime(__now)
#endif
      time_t now = time(NULL);
      struct tm res;
      strftime(buf, MAXSNLEN, "%Y%m%d-%H%M%S", localtime_r(&now, &res));
#ifdef _WIN32
#    undef localtime_r
#endif
    }

    if(Optmap.master_predel != NULL) {
        char cmd[MAXLNLEN];
        snprintf(cmd, MAXLNLEN, Optmap.master_predel,
         (unsigned int)ic, buf);
        system(cmd);
    }

    travp = dell->first;
    while(travp != NULL) {
        struct accdb_user rq;
        char *lname;

        memset(&rq, 0, sizeof(struct accdb_user));
        rq.lname = travp->ptr;
        rq.uid = rq.gid = -1;

        if((eax = Q_USERINFO(&rq, &rq, 1)) < 0) {
            printf("\n");
            fprintf(stderr, "userinfo: Something went wrong. (eax=%d, "
             "errno=%d: %s)\n", eax, errno, strerror(errno));
            return 0;
        } else if(eax == 0) {
            // This should not happen. Someone deleted the user before we did.
            printf("\n");
            fprintf(stderr, "Warning: Someone deleted \"%s\" before we did.\n",
             rq.lname);
            continue;
        }

        if(Optmap.user_predel != NULL) {
            char cmd[MAXLNLEN];
            snprintf(cmd, MAXLNLEN, Optmap.user_predel, rq.lname,
             rq.uid, rq.gid, rq.igrp, rq.home, buf);
            system(cmd);
        }
        HX_rrmdir(rq.home);

        // need a copy since travp->ptr will be undefined after userdel
        lname = HX_strdup(travp->ptr);

        // (surprise, we can reuse "rq" without modification)
        if((eax = Q_USERDEL(&rq)) < 0) {
            printf("\n");
            fprintf(stderr, "userdel: Something went wrong. (eax=%d, "
             "errno=%d: %s)\n", eax, errno, strerror(errno));
            return 0;
        }

        if(logp != NULL) { fprintf(logp, "-:%s\n", lname); }

        if(Optmap.user_postdel != NULL) {
            char cmd[MAXLNLEN];
            snprintf(cmd, MAXLNLEN, Optmap.user_postdel, lname);
            system(cmd);
        }

        free(lname);
        if((--ic & 0x3F) == 0) {
            printf("delete: %u users remaining\n", (unsigned int)ic);
        }
        travp = travp->Next;
    }

    printf("Successfully deleted %u users\n", (unsigned int)dell->itemcount);

    if(Optmap.master_postdel != NULL) {
        char cmd[MAXLNLEN];
        snprintf(cmd, MAXLNLEN, Optmap.master_postdel,
         (unsigned int)dell->itemcount, buf);
        system(cmd);
    }

    return 1;
}

void sync_free(struct HXbtree *addl, struct HXlld *dell) {
    if(addl != NULL) {
        void *travp = HXbtrav_init(addl, NULL);
        struct HXbtree_node *node;

        while((node = HXbtraverse(travp)) != NULL) {
            free_addl_entry(node->data);
        }

        HXbtrav_free(travp);
        HXbtree_free(addl);
    }

    if(dell != NULL) { HXlld_free(dell); }
    return;
}

//-----------------------------------------------------------------------------
inline static void free_addl_entry(struct ds_entry *e) {
    if(e == NULL) { return; }
    free(e->vname);
    free(e->nname);
    free(e->xuid);
    free(e->sgroup);
    free(e->lname);
    free(e->rname);
    free(e);
    return;
}

inline static void handle_user(struct accdb_user *p, struct HXbtree *addl,
 struct HXlld *dell, size_t *ukeep) {
    char buf[MAXLNLEN], *brkp;

    buf[MAXLNLEN - 1] = '\0';
    strncpy(buf, p->gecos, MAXLNLEN - 1);

    if((brkp = strchr(buf, ',')) != NULL &&
     (brkp = strchr(++brkp, ',')) != NULL) {
        // strip further fields
        *brkp = '\0';
    }

    if(HXbtree_find(addl, buf) != NULL) {
        ++*ukeep;
        free_addl_entry(HXbtree_get(addl, buf));
        HXbtree_del(addl, buf);
        // what remains in storage_addlist is the users to add
    } else {
        // not anymore in data source
        HXlld_push(dell, p->lname);
    }

    return;
}

//==[ End of file ]============================================================
